/* Copyright 2013 Sven Jacobs
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.svenjacobs.lugaene;

import com.googlecode.objectify.ObjectifyService;
import org.apache.lucene.store.Directory;
import org.apache.lucene.store.IOContext;
import org.apache.lucene.store.IndexInput;
import org.apache.lucene.store.IndexOutput;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Properties;

import static com.googlecode.objectify.ObjectifyService.ofy;

/**
 * Lucene {@link org.apache.lucene.store.Directory} that uses Google App Engine data store
 * <p/>
 * GaeDirectory makes it possible to run a Lucene index (or several) on GAE.
 * Just use Lucene as you would normally but instead of one of Lucene's default directories use GaeDirectory.
 * <p/>
 * Example:
 * <pre><code>
 * Directory directory = new GaeDirectory("MyIndex");
 * Analyzer analyzer = new StandardAnalyzer(Version.LUCENE_43);
 * IndexWriterConfig config = new IndexWriterConfig(Version.LUCENE_43, analyzer);
 * IndexWriter writer = new IndexWriter(directory, config);
 * ...
 * </code></pre>
 * <p/>
 * Specify <strong>unique</strong> names per directories at GaeDirectory's constructor when running multiple
 * directories on GAE.
 * <p/>
 * This library uses <a href="https://code.google.com/p/objectify-appengine/">Objectify</a> version 4 internally to
 * store entities on GAE. It has been tested with Lucene 4.3.0.
 *
 * @author Sven Jacobs
 */
public class GaeDirectory extends Directory {

    public static final String DEFAULT_NAME = "default";

    static {
        ObjectifyService.register(GaeDirectoryEntity.class);
        ObjectifyService.register(GaeFileEntity.class);
        ObjectifyService.register(GaeFileBufferEntity.class);
        ObjectifyService.register(GaeLockEntity.class);

        // Lucene tries to read the system properties os.arch and os.version however they are not available in GAE
        // runtime environment so we're setting them to some fake values.

        final Properties systemProps = System.getProperties();

        systemProps.setProperty("os.arch", "LuGAEne");
        systemProps.setProperty("os.version", "1.0");
    }

    private final String name;
    private GaeDirectoryEntity directoryEntity;

    /**
     * Creates an instance with the default name
     */
    public GaeDirectory() {
        this(DEFAULT_NAME);
    }

    /**
     * Creates an instance with the specified unique name
     *
     * @param name Unique name of this directory
     */
    public GaeDirectory(final String name) {
        this.name = name;

        try {
            setLockFactory(new GaeLockFactory(directoryEntity()));
        } catch (final IOException ignored) {
        }
    }

    @Override
    public String[] listAll() throws IOException {
        ensureOpen();

        final List<String> names = new ArrayList<String>();

        for (final GaeFileEntity file : ofy().load().type(GaeFileEntity.class).ancestor(directoryEntity())) {
            names.add(file.name);
        }

        return names.toArray(new String[names.size()]);
    }

    @Override
    public boolean fileExists(final String name) throws IOException {
        ensureOpen();
        return file(name) != null;
    }

    @Override
    public void deleteFile(final String name) throws IOException {
        ensureOpen();
        ofy().delete().type(GaeFileEntity.class).parent(directoryEntity()).id(name).now();
    }

    @Override
    public long fileLength(final String name) throws IOException {
        ensureOpen();

        final GaeFileEntity file = file(name);

        if (file == null) {
            throw new FileNotFoundException(name);
        }

        return file.size;
    }

    @Override
    public IndexOutput createOutput(final String name,
                                    final IOContext context) throws IOException {

        ensureOpen();
        return new GaeIndexOutput(new GaeFileEntity(directoryEntity(), name));
    }

    @Override
    public void sync(final Collection<String> names) throws IOException {
        // noop
    }

    @Override
    public IndexInput openInput(final String name,
                                final IOContext context) throws IOException {

        ensureOpen();

        final GaeFileEntity file = file(name);

        if (file == null) {
            throw new FileNotFoundException(name);
        }

        return new GaeIndexInput(file);
    }

    @Override
    public void close() throws IOException {
        isOpen = false;
    }

    private GaeFileEntity file(final String name) {
        return ofy().load().type(GaeFileEntity.class).parent(directoryEntity()).id(name).now();
    }

    private GaeDirectoryEntity directoryEntity() {
        if (directoryEntity != null) {
            return directoryEntity;
        }

        GaeDirectoryEntity entity = ofy().load().type(GaeDirectoryEntity.class).id(name).now();

        if (entity == null) {
            entity = new GaeDirectoryEntity(name);
            ofy().save().entity(entity).now();
        }

        directoryEntity = entity;

        return directoryEntity;
    }
}
